################################################################################
# CxInstrumentClass.tcl
# ------------------------------------------------------------------------------
# This script is provided as a utility supporting Paragon-neo and Paragon-100G.
# It is NOT intended as example code for controlling these instruments.
# No part of this script should be used independently of this script without
# the explicit written permission of Calnex Solutions Ltd.
################################################################################
source calnexRest.tcl
set err [catch {
	source CxRegisters.tcl
} res]

if {$err != 0} {
	cxLogDebug "CxRegisters.tcl: $res"
}

set err [catch {
	source CxRegisterOpticalModuleWrite.tcl
} res]

if {$err != 0} {
	cxLogDebug "CxRegisterOpticalModuleWrite.tcl: $res"
}

#source CxRegisters.tcl
source CxInstrumentOpticalModuleClass.tcl

oo::class create CxInstrumentEthernetPort {
	variable instrument
	variable PortName
	variable portNumber
	variable ipAddr
	
	constructor {instr portNum ipAddress} {
		set instrument $instr
		set portNumber $portNum
		set PortName Port$portNum
		set ipAddr $ipAddress
		
		cxLogDebug "[info level 0]"
		
		# HACK HACK
		#CxInstrumentOpticalModule create opticalModule [self] SFP28 $ipAddress

	}
	
	destructor {}

	method getInstrument {} {
		return $instrument
	}
	method getPortName {} {
		return $PortName
	}
	method getEthernetSettings {} {
		return [calnexGet physical/port/ethernet/$PortName]

	}
	method getSelectedInterface {} {
		return [calnexGetVal "physical/port/ethernet/$PortName" "Selected"]
	}
	method getSelectedInterfaceDetails {} {
		set retval [list "InterfaceType" "No Interface Selected"]
		set interface [my getSelectedInterface]
		if {$interface != "None"} {
			set retval [calnexGet "physical/port/ethernet/$PortName/$interface"]
		}
		return $retval
	}
	method getFecState {interface} {
		if {$interface != "QSFP28" && $interface != "SFP28" && $interface != "CFP4"} {
			return false
		}
		set retval [dict get [my getInterfaceDetails $interface] "Fec"]
		return $retval
	}
	method getInterfaceDetails {interface} {
		set retval [calnexGet "physical/port/ethernet/$PortName/$interface"]		
		return $retval		
	}
	method getFittedInterfaces {} {
		set fittedInterfaces {}
		set settings [my getEthernetSettings]
		set ifs [dict get $settings "Interfaces"]
		dict for {if ifInfo} $ifs {
			#set info [dict get $ifs $if]
			set fitted [dict get $ifInfo "Fitted"]
			if {$fitted} {
				lappend fittedInterfaces $if			
			}
		}
		return $fittedInterfaces
	}
	# For some reason the LED info is returned as an array not a dict.
	# We can't use calnexGetVal for this.
	method getLinkStatus {} {
		set num [expr $portNumber-1]
		set led "ethLink_$num"
		# Had to change calnexRest to be able to have a URL with ?LedNames=
		# TODO: calnexRest should support a better method for this. Need to fix...
		set ledArray [calnexGet "results/statusleds?LedNames=$led"]
		#puts $ledDetails
		set ledDetails [lindex $ledArray 0]
		set link [dict get $ledDetails "State"]
		return $link
	}
	
	forward opticalModule opticalModule
}

oo::class create CxInstrument {
	variable ipAddress
	variable serialNumber
	variable model
	variable capabilities
	variable softwareVersion
	variable osVersion
	variable restApi
	variable configBin
	variable configJson
	variable numSessionFolders
	variable dpePlusRegisters
    variable ptpOptionFitted
	#variable port1
	#variable port2
	
    constructor { ipAddr } {
		cxLogDebug "[info level 0]"

        set ipAddress $ipAddr
		set restApi "http://$ipAddress/api/"
		set configBin "method {get} headers \{Content-Type application/octet-stream\}"
		set configJson "method {get} headers \{Content-Type application/json\}"
		set numSessionFolders 0
		
		my Connect $ipAddr
		#calnexInit $ipAddress
		set serialNumber [calnexGetVal instrument/information SerialNumber]
		set model [calnexGetVal instrument/information HwType]
		#set capabilities [calnexGetVal instrument/information HwCapabilities]
		set capabilities "Not being reported correctly - ignore"
		set softwareVersion [calnexGetVal instrument/software/buildversion BuildVersion]
		set osVersion [calnexGetVal instrument/software/osversion OsVersion]
        
        set ptpOptionFitted [my IsPtpFitted]
		
		CxInstrumentEthernetPort create p1 self 1 $ipAddress
		CxInstrumentEthernetPort create p2 self 2 $ipAddress
		# If we have register access
		if {!([info commands CxRegister] == "")} {
			cxLogDebug "CxRegister exists"
			CxRegister create dpePlusRegisters "DPE_PLUS" "FPGA"
		} else {
			cxLogDebug "CxRegister does not exist"			
		}
	}

    destructor {}

	# Private method (starts with an uppercase character) 
	method Connect {ip} {
		set retries 1
		set retry 0
		set err 1
		
		while {$err != 0 && $retry < $retries} {
			cxLogDebug "Connect: calling calnexInit ($retry)"
			set err [catch {calnexInit $ip}]
			incr retry
		}
		if {$retry > $retries} {
			cxLogDebug "Unable to read instrument details. Please ensure that you can connect to the instrument from your browser."
			throw CONNECT_FAIL "Unable to read instrument details. Please ensure that you can connect to the instrument from your browser."
			#exit
		}
		
		# Switch off the script recorder
		calnexSet "/script/recorder/stop"
	}
	method getIpAddress {} {
		return $ipAddress
	}
	method getSerialNumber {} {
		return $serialNumber
	}
	method getModel {} {
		return $model
	}
	method getCapabilities {} {
		return $capabilities
	}
	method getSoftwareVersion {} {
		return $softwareVersion
	}
	method getOsVersion {} {
		return $osVersion
	}
	method getNumSessionFolders {} {
		return $numSessionFolders
	}
	method getOptions {} {
		return [calnexGet instrument/options/state]
	}
    method isPtpOptionFitted {} {
        return $ptpOptionFitted
    }
    method IsPtpFitted {} {
        set mseOptions {"NEO-PTP-G.8275.1" "NEO-PTP-G.8275.2" "250" "251" "252"}
        set options [my getOptions]
		foreach opt $options {
            set id [dict get $opt "OptId"]
            set fitted [dict get $opt "OptState"]
            if {$id in $mseOptions && $fitted == "Fitted"} {
                return "true"
            }
        }
        return "false"
    }
	method getAppStatus {} {
		return [calnexGet app/all]
	}
	method getCatStatus {} {
		return [calnexGet cat/general/status]
	}
	method getCatMeasurements {} {
		return [calnexGet cat/measurement]
	}
	method getAppStatusConformanceGeneration {} {
        if {$ptpOptionFitted} {
    		return [calnexGet app/conformance/generation]
        } else {
            return "PTP option not fitted"
        }
	}
	method getAppStatusConformanceMeasurement {} {
        if {$ptpOptionFitted} {
    		return [calnexGet app/conformance/measurement]
        } else {
            return "PTP option not fitted"
        }
	}
	method getAppStatusConformanceStimulus {} {
        if {$ptpOptionFitted} {
    		return [calnexGet app/conformance/stimulus]
        } else {
            return "PTP option not fitted"
        }
	}
	method getMseSetting {setting} {
        if {$ptpOptionFitted} {
    		return [calnexGet app/mse/$setting]
        } else {
            return "PTP option not fitted"
        }
	}
	method getMseMasterSetting {master setting} {
        if {$ptpOptionFitted} {
    		return [calnexGet app/mse/master/$master/$setting]
        } else {
            return "PTP option not fitted"
        }
	}
	method getMseMasterUnicastSlaves {master} {
        if {$ptpOptionFitted} {
    		return [calnexGet app/mse/master/$master/unicast/slaves]
        } else {
            return "PTP option not fitted"
        }
	}
	method getMseSlaveSetting {slave setting} {
        if {$ptpOptionFitted} {
    		return [calnexGet app/mse/slave/$slave/$setting]
        } else {
            return "PTP option not fitted"
        }
	}
	method getSystemTime {} {
		return "[calnexGetVal instrument/systemtime SystemTime]	[calnexGetVal instrument/systemtime SystemDate]"		
	}
	#method getPort1 {} {
	#	return port1
	#}
	#method getPort2 {} {
	#	return port2
	#}
	#method getEthernetPortSettings {port} {
	#	return [calnexGet physical/port/ethernet/$port]
	#}
	method getFrequencyReferenceSettings {} {
		return [calnexGet physical/references/in/clock]		
	}
	method getOnePpsReferenceSettings {} {
		return [calnexGet physical/references/in/onepps]		
	}
	method getLedStatus {} {
		return [calnexGet results/statusleds]				
	}
	method startMseCapture {} {
        if {$ptpOptionFitted} {
    		calnexSet "app/mse/measurement/start"
        } else {
    		cxLogDebug "Call to startMseCapture but no PTP option fitted"
        }
	}
	method stopMseCapture {} {
        if {$ptpOptionFitted} {
    		calnexSet "app/mse/measurement/stop"
        } else {
    		cxLogDebug "Call to startMseCapture but no PTP option fitted"
        }
	}
	method clearCaptureFilters {} {
		#dpePlusRegisters registerSet "0" "00202ec0" "65"		
		dpePlusRegisters registerSet2 "0" "0" "2109120" "101"		
	}
	method selectPreset {preset} {
		calnexSet instrument/preset Name "$preset"
	}
	method getCurrentPreset {} {
		return [calnexGetVal "instrument/preset" "Value"]
	}
	method getPortCouplingState {} {

		set err [catch {
			set state [calnexGetVal "physical/port/ethernet/group/P1P2/coupled" "Value"]
		} res]

		if {$err != 0} {
			cxLogDebug "Port coupling not supported"
			set state "Unsupported"
		}
		#set state [calnexGetVal "physical/port/ethernet/group/P1P2/coupled" "Value"]
		return $state
	}
	method anyMeasurementRunning {} {
		cxLogDebug "[info level 0]"

		set anyRunning FALSE
		set appAll [calnexGet app/all]
		#puts $appAll
		foreach app $appAll {
			#cxLogDebug $app
			set appName [dict get $app "AppRef"]
			set runState [dict get $app "RunState"]
			set err [catch {set isCapture [dict get $app "IsCapture"]} res]
			if {$err != 0} {set isCapture "true"}

			#puts "$appName is $runState"
			if {$runState != "Stopped" && $isCapture == "true"} {
				set anyRunning TRUE
			}
		}
		# There is no service (currently) that indicates whether MSE measurements
		# are running. Best work-around appears to be to check the time elapsed.
		# This is updated roughly once a second and so if it is moving, the
		# mneasurement is running; otherwise not.
#		if {$anyRunning == FALSE} {
#			set timeElapsedStart [my getMseElapsedTime]
#			for {set i 0} {$i < 6} {incr i} {
#				after 333
#				set timeElapsedNow [my getMseElapsedTime]
#				if {$timeElapsedNow != $timeElapsedStart} {
#					set anyRunning TRUE
#					break
#				}	
#			}
#		}
		# Check MSE separately
		if {$anyRunning == FALSE && $ptpOptionFitted} {
			set mseRunState [my getMseRunState]
			foreach item $mseRunState {
#				puts $item
				set name [dict get $item "Item"]
				set state [dict get $item "State"]
#				puts "$name $state"
				if {$name == "PacketCapture" && $state != "Stopped"} {
					set anyRunning TRUE
					break
				}
			}
		}		
		return $anyRunning
	}
	method getMseElapsedTime {} {
		return [calnexGetVal "app/mse/measurement/elapsedtime" "ElapsedTime"]
	}
	method getMseRunState {} {
        if {$ptpOptionFitted} {
    		return [calnexGetVal "app/mse/" "ItemRunState"]
        } else {
            return "STOPPED"
        }
	}
	method getLastSessionsFolder {} {
		#set resp [calnexGetVal "cat/filecommander/SessionsFolder" ChildrenElements]
		set resp [my getAllSessionFolders]
		#puts $resp
		set length [llength $resp]
		#puts $length
		#set folder [lindex $resp $length-1]
		set folder [lindex $resp 0]
		return $folder		
	}
	#
	# method getAllSessionFolders --
	# Arguments:
	# 	None
	# Results:
	#	Returns a list of Session folders on the instrument newest first.
	#	The instrument appears to return the list in alphabetic order with
	#	folders at the beginning and any files at the bottom. Note that
	#	alphabetic order for instrument generated folders is the same as oldest
	#	first given the date / time naming convention.
	# 	This method first strips any files from the list and then reverses the
	#	order so that the newest file is top of the list.
	method getAllSessionFolders {} {
		cxLogDebug "[info level 0]"

		set resp [calnexGetVal "cat/filecommander/SessionsFolder" ChildrenElements]
		set isFolder FALSE
		# Strip files
		while {!$isFolder} {
			set filename [lindex $resp end]
			set isFolder [calnexGetVal "cat/filecommander/SessionsFolder?FileId=$filename" IsDirectory]
			if {!$isFolder} {
				# Not a folder, so remove it from the list
				set resp [lrange $resp 0 end-1]
			}
		}
		# Now reverse the list so we have newest first
		set resp [lreverse $resp]
 		set numSessionFolders [llength $resp]
		return $resp		
	}

	method getAllReportFilenames {} {
		cxLogDebug "[info level 0]"

		set resp [calnexGetVal "cat/filecommander/ReportFolder" ChildrenElements]
		# Now reverse the list so we have newest first
		set resp [lreverse $resp]
 		set numReportFiles [llength $resp]
		return $resp		
	}
	
	# Download a file from the instrument
	# Parameters:
	#	folderType: If "SessionsFolder" then binary data; assume text otherwise
	#	srcFolder: The name of the folder on the instrument. For sessions files this is the name of the session folder e.g. Session_<date>
	#	file: The name of the capture file in the Session folder
	#	destFolder: The name of the folder on the local machines where the remote folder will be saved
	method downloadFile {folderType srcFolder file destFolder} {
		cxLogDebug "[info level 0]"
		
		set url "cat/filecommander/download/$folderType?AsAttachment=true&FileId=$srcFolder/$file"

		set f "$destFolder/$file"
		set fid [open $f "w"]
		if {$folderType == "SessionsFolder"} {
			fconfigure $fid -translation binary
			set config $configBin
		} else {
			set config $configJson			
		}
		
		set err [catch {
			puts $fid [rest::simple "${restApi}${url}" "" $config ""]
		} res]
		#puts $err

		if {$err != 0} {
			error "Unable to download $file from $srcFolder. Please check that this folder and file exist... Aborting."
			#exit
		}
		
		close $fid
		
	}
	
	# Download a capture file from the instrument
	# Parameters:
	#	srcFolder: The name of the Session folder e.g. Session_<data>
	#	file: The name of the capture in the Session folder
	#	destFolder: The name of the folder on the lcoal PC where the Session folder will be saved
	method downloadSessionFile {srcFolder file destFolder} {
		cxLogDebug "[info level 0]"
		set folderType "SessionsFolder"
		my downloadFile $folderType $srcFolder $file $destFolder
	}

	method downloadReportFile {file destFolder} {
		cxLogDebug "[info level 0]"
		set folderType "ReportFolder"
		my downloadFile $folderType "./" $file $destFolder
	}
	
	# Download all the captures files in a Session folder
	# Parameters:
	#	srcFolder: The name of the Session folder
	#	destFolder: The name of the folder on the local PC where the Session folder will be saved
	method downloadSessionFiles {srcFolder destFolder} {
		cxLogDebug "[info level 0]"
		set url "cat/filecommander/SessionsFolder?FileId=$srcFolder&format=json"
		
		#set files [calnexGetVal  "cat/filecommander/SessionsFolder?FileId=$folder" ChildrenElements]
		set err [catch {
			set resp [rest::simple "${restApi}${url}" "" $configJson ""]			
		} res]
		#puts $err
		
		if {$err != 0} {
			error "Unable to access $srcFolder. Please check that this folder exists."
		}
		
		set files [dict get [json::json2dict $resp] ChildrenElements]
		#puts $files
		foreach file $files {
			my downloadSessionFile $srcFolder $file $destFolder			
		}
	}
	method downloadLogFiles {srcFolder destFolder} {
		cxLogDebug "[info level 0]"
		set folderType "LogsFolder"
		set url "cat/filecommander/$folderType?FileId=$srcFolder&format=json"
		
		#set files [calnexGetVal  "cat/filecommander/SessionsFolder?FileId=$folder" ChildrenElements]
		set err [catch {
    		set resp [rest::simple "${restApi}${url}" "" $configJson ""]
		} res]

        if {$err == 0} {
    		set files [dict get [json::json2dict $resp] ChildrenElements]
    		#puts $files
    		foreach file $files {
				
				##### THIS IS A HACK TO STOP THE ISSUE WITH DOWNLOADING THE NESTED FOLDER "MemDiagnostics"
				# It should be sorted to detect and properly download nested folders 
				# There is a field property "isFolder" or similar
				#
				if {$file == "MemDiagnostics"} {
					continue
				}
		###################################################
			my downloadFile $folderType $srcFolder $file $destFolder
    						
    		}
        } else {
            cxLogDebug "WARNING: $srcFolder may not exist on the instrument"
       		set f "$destFolder/$srcFolder folder does not appear to exiot on the instrument"
    		set fid [open $f "w"]
            close $fid

        }
	}
	method downloadLogs {logFilesFolder embeddedLogFilesFolder controlFolder} {
		my downloadLogFiles LogFiles $logFilesFolder
		my downloadLogFiles EmbeddedLogs $embeddedLogFilesFolder
		my downloadLogFiles Control $controlFolder
	}

	method pfvShow {} {
		eval exec [auto_execok start] \"\" [list "http://$ipAddress/PFV/#screen=resultsScreen"]
		return
	}

	forward port1 p1
	forward port2 p2

}

